import java.util.Formatter; /* This import is needed by 
                               toString, see end of file */

/** Data structure that allows potentially infinite long list of integers
 *  Similar to RList from CS61A.
 *  @author Josh Hug (with lots of Paul Hilfinger code at the end)
 */

public class IntList {
    int head;     /* similar to car of a list */
    IntList tail; /* similar to cdr of a list */

    public IntList(int head, IntList tail) {
        this.head = head;
        this.tail = tail;
    }

    /** increments P by N non-destructively, returning a new list */
    public static IntList incrList(IntList P, int n) {
        if (P == null)
            return null;

        /* List is not null, so make a new list that is incremented */
        
        return new IntList(P.head + n, incrList(P.tail, n));
    }

    /** increments P by N non-destructively, returning a new list */
    public static IntList incrListIterative(IntList P, int n) {
        if (P == null)
            return null;

/* used this code to get ourselves thinking!

        IntList R = new IntList(P.head + n, null);
        P = P.next;
        R.tail = new IntList(P.head + n, null);
        P = P.next;
        //uhoh! Lost the return array!
        R.tail = new IntList(P.head + n, null);
        R = R.tail;

        */

        /* this was our second exploration
        IntList R = new IntList(P.head + n, null);
        IntList rp = R;

        P = P.next;

        rp.tail = new IntList(P.head + n, null);
        rp = rp.tail;
        */

/*  This was our 3rd pass, in which we got it working.
        IntList R = null;
        IntList rp = null;

        while (P != null) {
            if (R == null) {
                R = new IntList(P.head + n, null);
                rp = R;
            }
            else {
                rp.tail = new IntList(P.head + n, null);
                rp = rp.tail;
            }
            P = P.tail;
        }
        return R;
        */


        /* This was our final pass, which we simplified
        by ensuring that the IntList was never empty when
        the loop began */

        IntList pp = P;
        IntList R = new IntList(P.head + n, null);
        IntList rp = R;
        while (pp.tail != null) {
            /* R points at a list of all items 
               to the left of pp, but + n 

               rp points at the last thing in R*/
            pp = pp.tail;
            rp.tail = new IntList(pp.head + n, null);
            rp = rp.tail;
        } 

        return R;
    }



    public static void main(String[] args) {
        IntList x = IntList.list(0, 1, 2);
        IntList xp2 = IntList.list(2, 3, 4);
        IntList actual = IntList.incrListIterative(x, 2);
    }



    /** increments P by N destructively, do not use new!
     *  Using new would defeat the point of this exercise!
     */
    public static IntList dincrList(IntList P, int n) {
        if (P == null)
            return null;

        P.head += n;
        /* P.tail = could be removed for this operation.
           Recursive operations that add/remove boxes
           will need this. */

        P.tail = dincrList(P.tail, n);
        return P;
    }

   /** increments P by N destructively, do not use new!
     *  Using new would defeat the point of this exercise!
     */
    public static IntList dincrListIterative(IntList P, int n) {
        IntList pp = P;
        while (pp != null) {
            pp.head += n;
            pp = pp.tail;
        }
        return P;
    }    


   /** recursively non-destructively removes instances of X from P
     */
    public static IntList removeAll(IntList P, int x) {
        if (P == null)
            return null;
        if (P.head == x)
            return removeAll(P.tail, x);
        return new IntList(P.head, removeAll(P.tail, x));
    }    


   /** iteratively non-destructively removes instances of X from P
     */
    public static IntList removeAllIterative(IntList P, int x) {
        if (P == null)
            return null;

        IntList R = null;
        IntList rp = R;

        /* R points to a list of all items to the left of 
           pp that are not equal to n.

           rp points at the end of R
        */
        /* before loop starts, is this true?
         * 
         * pp points at leftmost item, so R is correct
         * rp points at the last item of R (null, so correct)
        */
        for (IntList pp = P; pp != null; pp = pp.tail) {
            if (pp.head == x) {
                continue; /* new keyword, means jump to end of loop */
            }
            if (R == null) {
                R = new IntList(pp.head, null);
                rp = R;
            }
            else {
                rp.tail = new IntList(pp.head, null);
                rp = rp.tail;
            }
        }
        return R;
    }    


   /** recursively destructively removes instances of X from P
     */
    public static IntList dremoveAll(IntList P, int x) {
        if (P == null)
            return null;
        if (P.head == x)
            return dremoveAll(P.tail, x);
        P.tail = dremoveAll(P.tail, x);
        return P;
    }    



   /** iterative destructively removes instances of X from P
     */

    /* Remove all xs from L, destructive. don't use new */
    public static IntList dremoveAllIterative(IntList P, int x) {


        IntList rp = P;
        
        /*find first non-n entry of P */
        while ((rp != null) && (rp.head == x))
            rp = rp.tail;

        for (IntList pp = P; pp != null; ) {
            IntList next = pp.tail;
            if (next == null)
                break; /* done processing list */
            if (next.head == x) {
                pp.tail = next.tail;
            }
            else {
                pp = pp.tail;
            }
        }
        return rp;
    }


   /* Remove all xs from L, destructive. don't use new */
    public static IntList dremoveAllIterativeRefactored(IntList P, int x) {

        /* Observations:
           Might return a different list than the one we started with!

           We'll need:
           . Pointer to head of new list to return.

        */

        IntList rp = P;
        
        /*find first non-n entry of P */
        while ((rp != null) && (rp.head == x))
            rp = rp.tail;

        IntList last = null; 
        for (IntList pp = P; pp != null; ) {
            IntList next = pp.tail;
            if (pp.head != x) {
                if (rp == null) {
                    rp = pp;
                    last = pp;
                } else {
                    last.tail = pp;
                    last = pp;
                }
                pp.tail = null;
            }
            pp = pp.tail;
        }

        return rp;
    }


    /****** BELOW THIS LINE ARE HELPFUL METHODS FROM THE
            IntList CLASS FROM HW1. THESE WERE NOT DEVELOPED
            IN CLASS! ********/

    /** Returns a new IntList containing the ints in ARGS. */
    public static IntList list(Integer ... args) {
        IntList result, p;

        if (args.length > 0) {
            result = new IntList(args[0], null);
        } else {
            return null;
        }

        int k;
        for (k = 1, p = result; k < args.length; k += 1, p = p.tail) {
            p.tail = new IntList(args[k], null);
        }
        return result;
    }


    @Override
    public String toString() {
        Formatter out = new Formatter();
        String sep;
        sep = "(";
        int cnt = 0;

        for (IntList p = this; p != null; p = p.tail) {
            out.format("%s%d", sep, p.head);
            sep = ", ";

            cnt++;

        }
        out.format(")");
        return out.toString();
    }



    /** Returns true iff X is an IntList containing the same sequence of ints
     *  as THIS. Cannot handle IntLists with cycles. */
    public boolean equals(Object x) {
        if (!(x instanceof IntList)) {
            return false;
        }
        IntList L = (IntList) x;
        IntList p;
        
        for (p = this; p != null && L != null; p = p.tail, L = L.tail) {
            if (p.head != L.head) {
                return false;
            }
        }
        if (p != null || L != null) {
            return false;
        }
        return true;
    }

} 